
package de.codedo.conflicteditor;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.codehaus.jackson.JsonNode;

public class DocumentMatcher
{
    // @formatter:off
    private static final Collection<String> IGNORED_FIELD_NAMES = Arrays.asList(
        "_id", "_rev","_conflicts"
    );
    // @formatter:on

    private JsonNode _original;
    private JsonNode _other;
    private List<Difference> _differences;

    public DocumentMatcher(JsonNode original, JsonNode other)
    {
        super();
        _original = original;
        _other = other;
    }

    public List<Difference> compare()
    {
        if (_differences == null)
        {
            _differences = new ArrayList<Difference>();
            createDifferencesFromFieldsOnlyAvailableInOriginal();
            createDifferencesFromFieldsOnlyAvailableInConflict();
            createDifferencesFromFieldValues();
        }
        return _differences;
    }

    private void createDifferencesFromFieldsOnlyAvailableInOriginal()
    {
        Set<String> originalFieldNames = getFieldNames(_original);
        Set<String> otherFieldNames = getFieldNames(_other);
        originalFieldNames.removeAll(otherFieldNames);

        for (String name : originalFieldNames)
        {
            // TODO value may not be text
            String value = _original.findValue(name).getTextValue();
            Difference diff = new Difference(name, value, null);
            _differences.add(diff);
        }
    }

    private void createDifferencesFromFieldsOnlyAvailableInConflict()
    {
        Set<String> originalFieldNames = getFieldNames(_original);
        Set<String> otherFieldNames = getFieldNames(_other);
        otherFieldNames.removeAll(originalFieldNames);

        for (String name : otherFieldNames)
        {
            // TODO value may not be text
            String value = _other.findValue(name).getTextValue();
            Difference diff = new Difference(name, null, value);
            _differences.add(diff);
        }
    }

    private void createDifferencesFromFieldValues()
    {
        Set<String> originalFieldNames = getFieldNames(_original);
        Set<String> otherFieldNames = getFieldNames(_other);
        originalFieldNames.retainAll(otherFieldNames);

        for (String name : originalFieldNames)
        {
            Difference difference = compareValuesForKey(name);
            if (difference != null)
            {
                _differences.add(difference);
            }
        }
    }

    private Difference compareValuesForKey(String key)
    {
        JsonNode original = _original.findValue(key);
        JsonNode other = _other.findValue(key);

        if ((original != null) && (other == null))
        {
            // TODO this may be something different than plain text, e.g. some structured object
            return new Difference(key, original.getValueAsText(), null);
        }
        else if ((original == null) && (other != null))
        {
            // TODO this may be something different than plain text, e.g. some structured object
            return new Difference(key, null, other.getValueAsText());
        }
        else
        {
            return compareValues(key, original, other);
        }
    }

    private Difference compareValues(String key, JsonNode original, JsonNode other)
    {
        Object originalValue = original.getValueAsText();
        Object otherValue = other.getValueAsText();
        if (originalValue.equals(otherValue) == false)
        {
            return new Difference(key, originalValue, otherValue);
        }
        return null;
    }

    private Set<String> getFieldNames(JsonNode node)
    {
        Set<String> fieldNames = new HashSet<String>();
        Iterator<String> fieldNameIter = node.getFieldNames();
        while (fieldNameIter.hasNext())
        {
            String name = fieldNameIter.next();
            if (IGNORED_FIELD_NAMES.contains(name) == false)
            {
                fieldNames.add(name);
            }
        }
        return fieldNames;
    }
}
